
#include <Windows.h>
#include <Python.h>

#include <Winternl.h>

/*
 * The declaration of RTL_USER_PROCESS_PARAMETERS in Winternl.h is lacking.
 */
typedef struct {
    ULONG MaximumLength;
    ULONG Length;
    ULONG Flags;
    ULONG DebugFlags;
    PVOID ConsoleHandle;
    ULONG ConsoleFlags;
    PVOID StandardInput;
    PVOID StandardOutput;
    PVOID StandardError;
    UNICODE_STRING CurrentDirectoryPath;
    HANDLE CurrentDirectoryHandle;
    UNICODE_STRING DllPath;
    UNICODE_STRING ImagePathName;
    UNICODE_STRING CommandLine;
} MY_USER_PROCESS_PARAMETERS;

#define COMPILE_ASSERT(x) extern int dummy[(int)x]

COMPILE_ASSERT(sizeof(MY_USER_PROCESS_PARAMETERS) == sizeof(RTL_USER_PROCESS_PARAMETERS));


typedef
NTSTATUS
(NTAPI *QueryInformationProcess) (
    IN HANDLE ProcessHandle,
    IN PROCESSINFOCLASS ProcessInformationClass,
    OUT PVOID ProcessInformation,
    IN ULONG ProcessInformationLength,
    OUT PULONG ReturnLength OPTIONAL
    );

typedef ULONG (NTAPI *NtStatusToDosError) (NTSTATUS Status);

static QueryInformationProcess MyQueryInformationProcess;
static NtStatusToDosError      MyNtStatusToDosError;


static void *
objectAsHandle(PyObject *obj)
{
    PyObject *intobj;
    void *res;

    intobj = PyNumber_Int(obj);
    if (intobj == NULL) {
        return NULL;
    }
    res = PyLong_AsVoidPtr(intobj);
    Py_DECREF(intobj);
    return res;
}


static int
fireCallback(PyObject *callback, DWORD pid)
{
    PyObject *pidobj, *res;

    pidobj = PyLong_FromUnsignedLong(pid);
    if (pidobj == NULL) {
        return 0;
    }
    res = PyObject_CallFunctionObjArgs(callback, pidobj, NULL);
    Py_DECREF(pidobj);
    if (res == NULL) {
        return 0;
    }
    Py_DECREF(res);
    return 1;
}

static int
closeHandleChecked(HANDLE h)
{
    if (h != NULL && !CloseHandle(h)) {
        PyErr_SetFromWindowsErr(0);
        return 0;
    }
    return 1;
}


static PyObject *
runDebuggerMainloop(PyObject *self, PyObject *args)
{
    int starting = 1;
    long active = 0;
    HANDLE thandle;
    PyObject *thread, *processStarted, *processEnded;

    if (!PyArg_ParseTuple(args, "OOO:runDebuggerMainloop",
                          &thread, &processStarted, &processEnded)) {
        return NULL;
    }
    thandle = objectAsHandle(thread);
    if (thandle == NULL) {
        return NULL;
    }
    if (!PyCallable_Check(processStarted)) {
        PyErr_SetString(PyExc_TypeError, "'processStarted' must be callable");
        return NULL;
    }
    if (!PyCallable_Check(processEnded)) {
        PyErr_SetString(PyExc_TypeError, "'processEnded' must be callable");
        return NULL;
    }
    if (ResumeThread(thandle) == (DWORD)-1) {
        return PyErr_SetFromWindowsErr(0);
    }
    while (starting || active > 0) {
        DEBUG_EVENT event;

        if (!WaitForDebugEvent(&event, INFINITE)) {
            return PyErr_SetFromWindowsErr(0);
        }
        switch (event.dwDebugEventCode) {
            case CREATE_PROCESS_DEBUG_EVENT:
                starting = 0;
                active++;
                if (!closeHandleChecked(event.u.CreateProcessInfo.hFile)) {
                    return NULL;
                }
                if (!fireCallback(processStarted, event.dwProcessId)) {
                    return NULL;
                }
                break;
            case EXIT_PROCESS_DEBUG_EVENT:
                active--;
                if (!fireCallback(processEnded, event.dwProcessId)) {
                    return NULL;
                }
                break;
            /*
             * mspdbsrv.exe goes wild when this leak is plugged,
             * configure breaks if it isn't.
             */
            case LOAD_DLL_DEBUG_EVENT:
                if (!closeHandleChecked(event.u.LoadDll.hFile)) {
                    return NULL;
                }
                break;
        }
        if (!ContinueDebugEvent(event.dwProcessId,
                                event.dwThreadId,
                                DBG_EXCEPTION_NOT_HANDLED)) {
            return PyErr_SetFromWindowsErr(0);
        }
    }
    Py_RETURN_NONE;
}


static PyObject *
readProcessParameter(HANDLE phandle, UNICODE_STRING *param)
{
    PyObject *res;
    wchar_t *buf = malloc(param->Length);

    if (buf == NULL) {
        return PyErr_NoMemory();
    }
    if (!ReadProcessMemory(phandle, param->Buffer,
                            buf, param->Length, NULL)) {
        free(buf);
        return PyErr_SetFromWindowsErr(0);
    }
    res = PyUnicode_FromWideChar(buf, param->Length / sizeof(wchar_t));
    free(buf);
    return res;
}


static PyObject *
queryInformationProcess(PyObject *self, PyObject *process)
{
    HANDLE phandle;
    PROCESS_BASIC_INFORMATION info;
    NTSTATUS status;
    PEB envBlock;
    MY_USER_PROCESS_PARAMETERS params;
    PyObject *exe, *cmdline, *cwd;

    phandle = objectAsHandle(process);
    if (phandle == NULL) {
        return NULL;
    }
    status = MyQueryInformationProcess(phandle, ProcessBasicInformation,
                                       &info, sizeof(info), NULL);
    if (!NT_SUCCESS(status)) {
        SetLastError(MyNtStatusToDosError(status));
        return PyErr_SetFromWindowsErr(0);
    }
    if (!ReadProcessMemory(phandle, info.PebBaseAddress,
                           &envBlock, sizeof(envBlock), NULL)) {
        return PyErr_SetFromWindowsErr(0);
    }
    if (!ReadProcessMemory(phandle, envBlock.ProcessParameters,
                           &params, sizeof(params), NULL)) {
        return PyErr_SetFromWindowsErr(0);
    }
    exe = readProcessParameter(phandle, &params.ImagePathName);
    if (exe == NULL) {
        return NULL;
    }
    cmdline = readProcessParameter(phandle, &params.CommandLine);
    if (cmdline == NULL) {
        Py_DECREF(exe);
        return NULL;
    }
    cwd = readProcessParameter(phandle, &params.CurrentDirectoryPath);
    if (cwd == NULL) {
        Py_DECREF(exe);
        Py_DECREF(cmdline);
        return NULL;
    }
    return Py_BuildValue("nNNN", info.Reserved3, exe, cmdline, cwd);
}


static int
timePointToMilliseconds(FILETIME *time, unsigned PY_LONG_LONG *result)
{
    SYSTEMTIME st;
    struct tm tm = {0};
    time_t t;

    if (!FileTimeToSystemTime(time, &st)) {
        PyErr_SetFromWindowsErr(0);
        return 0;
    }
    tm.tm_year = st.wYear - 1900;
    tm.tm_mon = st.wMonth - 1;
    tm.tm_mday = st.wDay;
    tm.tm_hour = st.wHour;
    tm.tm_min = st.wMinute;
    tm.tm_sec = st.wSecond;
    tm.tm_year = st.wYear - 1900;
    tm.tm_isdst = -1;
    t = mktime(&tm);
    if (t == (time_t)-1) {
        PyErr_SetFromErrno(PyExc_OSError);
        return 0;
    }
    *result = t * 1000 + st.wMilliseconds;
    return 1;
}


static unsigned PY_LONG_LONG
timePeriodToMilliseconds(FILETIME time)
{
    ULARGE_INTEGER i;

    i.LowPart = time.dwLowDateTime;
    i.HighPart = time.dwHighDateTime;
    return i.QuadPart / 10000;
}


static PyObject *
getProcessTimes(PyObject *self, PyObject *process)
{
    FILETIME start, stop, kernel, user;
    unsigned PY_LONG_LONG startms, stopms, kernelms, userms;
    HANDLE phandle;

    phandle = objectAsHandle(process);
    if (phandle == NULL) {
        return NULL;
    }
    if (!GetProcessTimes(phandle, &start, &stop, &kernel, &user)) {
        return PyErr_SetFromWindowsErr(0);
    }
    if (!timePointToMilliseconds(&start, &startms) ||
        !timePointToMilliseconds(&stop, &stopms)) {
        return NULL;
    }
    kernelms = timePeriodToMilliseconds(kernel);
    userms = timePeriodToMilliseconds(user);
    return Py_BuildValue("KKKK", startms, stopms, kernelms, userms);
}


static PyMethodDef mwtraceFunctions[] = {
    {"runDebuggerMainloop",  runDebuggerMainloop, METH_VARARGS, NULL},
    {"queryInformationProcess", queryInformationProcess, METH_O, NULL},
    {"getProcessTimes", getProcessTimes, METH_O, NULL},
    {NULL, NULL, 0, NULL} /* Sentinel */
};


PyMODINIT_FUNC
init_mwtrace(void)
{
    HMODULE ntdll = LoadLibrary(TEXT("ntdll.dll"));

    if (ntdll == NULL) {
        PyErr_SetFromWindowsErr(0);
        return;
    }
    MyQueryInformationProcess = (QueryInformationProcess)GetProcAddress(ntdll, "NtQueryInformationProcess");
    MyNtStatusToDosError = (NtStatusToDosError)GetProcAddress(ntdll, "RtlNtStatusToDosError");
    if (MyQueryInformationProcess == NULL || MyNtStatusToDosError == NULL) {
        PyErr_SetFromWindowsErr(0);
        return;
    }
    (void)Py_InitModule("_mwtrace", mwtraceFunctions);
}
