
"""
Collect information about a tree of processes such as
process run times, command line, memory and I/O statistics.
"""

import argparse
import itertools
import json
import os
import sys
import subprocess
import time

import win32api
import win32con
import win32event
import win32file
import win32process
import win32profile

import _mwtrace



class Tracer(object):


    def __init__(self, argv, env=None, profileDir=None):
        self.argv = argv
        self.env = env
        self.profileDir = profileDir


    def run(self):
        self.seen = {}
        self.running = {}
        self.exiting = {}

        commandLine = subprocess.list2cmdline(self.argv)
        startupInfo = self.prepareStartupInfo()

        phandle, thandle, pid, tid = win32process.CreateProcess(
            None,                           # appName
            commandLine,
            None,                           # processAttributes
            None,                           # threadAttributes
            True,                           # inheritHandles
            win32process.DEBUG_PROCESS |
            win32process.CREATE_SUSPENDED,
            self.env,
            None,                           # cwd
            startupInfo)

        try:
            _mwtrace.runDebuggerMainloop(
                thandle,
                self.processStarting,
                self.processEnding)
        finally:
            phandle.close()
            thandle.close()
            self.cleanupStartupInfo(startupInfo)

        while self.exiting:
            self.reapProcesses(win32event.INFINITE)
        return [process.info for process in self.seen.itervalues()]


    def processStarting(self, pid):
        process = Process(pid)
        self.running[process.pid] = process
        self.seen[process.unique] = process
        parent = self.running.get(process.ppid)
        if parent is not None:
            process.info["ParentProcessId"] = parent.unique
        self.processStarted(process)


    def processEnding(self, pid):
        process = self.running.pop(pid, None)
        if process is not None:
            self.exiting[process.handle] = process
            pdir = self.profileDir
            if pdir is not None:
                old = os.path.join(pdir, "%s.pstats" % (pid,))
                new = os.path.join(pdir, "process%s.pstats" % (process.unique),)
                if os.path.exists(old):
                    try:
                        os.rename(old, new)
                    except OSError:
                        pass
                    else:
                        process.info["ProfileFile"] = new
        self.reapProcesses(0)


    def reapProcesses(self, initialTimeout):
        handles = self.exiting.keys()
        timeout = initialTimeout
        while handles:
            limited = handles[:win32event.MAXIMUM_WAIT_OBJECTS]
            result = win32event.WaitForMultipleObjects(limited, False, timeout)
            if result == win32event.WAIT_TIMEOUT:
                handles = handles[win32event.MAXIMUM_WAIT_OBJECTS:]
            else:
                index = result - win32event.WAIT_OBJECT_0
                handle = handles[index]
                process = self.exiting.pop(handle)
                handles = handles[index + 1:]
                process.processEnded()
                self.processEnded(process)
            timeout = 0


    def processStarted(self, process):
        pass


    def processEnded(self, process):
        pass



    def prepareStartupInfo(self):
        info = win32process.STARTUPINFO()
        info.dwFlags = win32process.STARTF_USESTDHANDLES
        info.hStdInput = self.inheritableStdHandle(win32api.STD_INPUT_HANDLE)
        info.hStdOutput = self.inheritableStdHandle(win32api.STD_OUTPUT_HANDLE)
        info.hStdError = self.inheritableStdHandle(win32api.STD_ERROR_HANDLE)
        return info


    def cleanupStartupInfo(self, info):
        info.hStdInput.close()
        info.hStdOutput.close()
        info.hStdError.close()


    def inheritableStdHandle(self, whichOne):
        handle = win32api.GetStdHandle(whichOne)
        return self.inheritableHandle(handle)


    def inheritableHandle(self, handle):
        if handle != win32file.INVALID_HANDLE_VALUE:
            handle = win32api.DuplicateHandle(
                win32api.GetCurrentProcess(),
                handle,
                win32api.GetCurrentProcess(),
                0,          # desiredAccess
                True,       # inheritHandle
                win32con.DUPLICATE_SAME_ACCESS)
        return handle



class Process(object):

    _nextId = iter(itertools.count(1)).next


    def __init__(self, pid):
        self.handle = win32api.OpenProcess(
            win32con.PROCESS_QUERY_INFORMATION |
            win32con.PROCESS_VM_READ |
            win32con.SYNCHRONIZE,
            False, # inheritHandle
            pid)
        ppid, exe, cmdline, cwd = _mwtrace.queryInformationProcess(self.handle)
        self.info = {}
        self.pid  = pid
        self.ppid = ppid
        self.unique = self.info["ProcessId"] = self._nextId()
        self.info["ParentProcessId"] = 0
        self.info["ImagePathName"] = exe.encode("utf8")
        self.info["CommandLine"] = cmdline.encode("utf8")
        self.info["InitialWorkingDirectory"] = cwd.encode("utf8")


    def processEnded(self):
        self.info["ExitCode"] = win32process.GetExitCodeProcess(self.handle)
        self.info.update(win32process.GetProcessIoCounters(self.handle))
        memoryInfo = win32process.GetProcessMemoryInfo(self.handle)
        # We are only interested in peak values
        for name in ("PagefileUsage", "QuotaNonPagedPoolUsage",
                     "QuotaPagedPoolUsage", "WorkingSetSize"):
            del memoryInfo[name]
        self.info.update(memoryInfo)
        start, stop, kernel, user = _mwtrace.getProcessTimes(self.handle)
        self.info["CreationTime"] = start
        self.info["ExitTime"] = stop
        self.info["KernelTime"] = kernel
        self.info["UserTime"] = user
        self.info["RealTime"] = stop - start
        self.handle.close()



def _abspath(path):
    return os.path.abspath(unicode(path, sys.getfilesystemencoding()))


def _setupProfiling(profileDir):
    if profileDir is None:
        return None, None
    profileDir = _abspath(profileDir)
    if os.path.exists(profileDir):
        raise RuntimeError("PROFILEDIR parameter %r already exists" % profileDir)
    else:
        os.makedirs(profileDir)
    env = win32profile.GetEnvironmentStrings()
    env.pop(u"PYTHONNOUSERSITE", None)
    installDir = _abspath(os.path.dirname(_mwtrace.__file__))
    env[u"PYTHONUSERBASE"] = os.path.join(installDir, u"userbase")
    env[u"MWTRACE_PROFILE_DIR"] = profileDir
    exe = _abspath(sys.executable)
    drive, rest = os.path.splitdrive(exe)
    python = u"/" + drive[0].lower() + rest.replace("\\", "/")
    env[u"PYTHON"] = python
    return env, profileDir


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-o", "--output", default="<stdout>", metavar="OUTFILE",
        help="store collected information in file named OUTFILE (default: stdout)")
    parser.add_argument("-p", "--profile", default=None, metavar="PROFILEDIR",
        help="enable Python profiling and store stats for each individual "
             "process in a separate file in PROFILEDIR")
    parser.add_argument("program", metavar="path\\to\\program",
        help="path to the program that should be traced")
    parser.add_argument("args", nargs=argparse.REMAINDER, metavar="...",
        help="arguments that should be passed to the traced program")
    options = parser.parse_args()
    env, profileDir = _setupProfiling(options.profile)

    tracer = Tracer([options.program] + options.args, env, profileDir)
    processes = tracer.run()

    try:
        console = open("CONOUT$", "w")
    except IOError:
        # not connected to a console
        pass
    else:
        print >> console,  "=================================================="
        print >> console, ("Traced %d processes. Writing details to '%s'." %
                           (len(processes), options.output))
        console.flush()

    if options.output in ("-", "<stdout>"):
        file = sys.stdout
    else:
        file = open(options.output, "wb")
    json.dump(processes, file, indent=2, separators=(",",": "), sort_keys=True)


if __name__ == "__main__":
    main()
