
import ctypes
import json
import operator
import os
import pstats
import sys

from keyword import iskeyword
from cStringIO import StringIO


PYTHON_OPTIONS = set(u"-" + o for o in "B E O OO Q R s S t u x 3 U".split())


def main(dirname):
    infopath = os.path.join(dirname, "info.json")
    if os.path.exists(infopath):
        with open(infopath) as f:
            info = json.load(f)
    else:
        info = {}
    return report(dirname, info, *analyze(dirname))


def analyze(dirname):
    dirname = os.path.abspath(dirname)
    processes = {}
    programs = {}

    with open(os.path.join(dirname, "mwtrace.json")) as infile:
        for process in json.load(infile):
            program = os.path.basename(process["ImagePathName"]).lower()
            argv = CommandLineToArgv(process["CommandLine"])
            if program == "python.exe":
                i = 1
                while argv[i] in PYTHON_OPTIONS:
                    i += 1
                if argv[i] == "-m":
                    i += 1
                elif argv[i] == "-c":
                    i += 1
                    stmt = argv[i].split("\n")[0].split(";")[0]
                    argv[i] = ":".join(s for s in stmt.split()
                                       if not iskeyword(s))
                elif os.path.basename(argv[i]) == "pythonpath.py":
                    i += 1
                    while argv[i].startswith("-I"):
                        if argv[i] == "-I":
                            i += 2
                        else:
                            i += 1
                program = os.path.basename(argv[i])
            process["Application"] = program
            process["children"] = []
            programs.setdefault(program, []).append(process)
            processes[process["ProcessId"]] = process

    processCount = len(processes)
    roots = []
    dummy = {"children": roots}
    for process in processes.values():
        parent = processes.get(process["ParentProcessId"], dummy)
        parent["children"].append(process)

    assert len(roots) == 1
    root = roots[0]

    ctime = operator.itemgetter("CreationTime")
    def ctime(process):
        return process["CreationTime"]

    for process in processes.itervalues():
        process["children"].sort(key=ctime)

    programList = []
    for program, procs in programs.iteritems():
        info = {}
        info["Application"] = program
        info["Invocations"] = len(procs)
        statsfn = program.replace(":", ".") + ".pstats"
        statspath = os.path.join(dirname, statsfn)
        if os.path.exists(statspath):
            stats = pstats.Stats(statspath)
        else:
            stats = None
        for process in procs:
            for measurement in ("KernelTime", "UserTime", "PageFaultCount",
                                "OtherOperationCount", "OtherTransferCount",
                                "ReadOperationCount", "ReadTransferCount",
                                "WriteOperationCount", "WriteTransferCount"):
                info[measurement] = (info.get(measurement, 0) +
                                     process[measurement])
            for measurement in ("PeakPagefileUsage", "PeakWorkingSetSize"):
                info[measurement] = max(info.get(measurement, 0),
                                        process[measurement])
            pfile = process.get("ProfileFile")
            if pfile is not None and os.path.exists(pfile):
                if os.path.getsize(pfile) == 0:
                    parent = processes[process["ParentProcessId"]]
                    print "empty profile for '%s':" % process["Application"]
                    print "     parent command line:", parent["CommandLine"]
                    print "       working directory:", process["InitialWorkingDirectory"]
                elif stats is None:
                    stats = pstats.Stats(pfile)
                else:
                    stats.add(pfile)
                os.remove(pfile)
        if stats is not None:
            stats.dump_stats(statspath)
            info["ProfileFileName"] = statsfn
        programList.append(info)
    return len(processes), root, programList


def CommandLineToArgv(commandLine):
    size = ctypes.c_int()
    ptr = ctypes.windll.shell32.CommandLineToArgvW(commandLine.decode("utf8"),
                                                   ctypes.byref(size))
    ref = ctypes.c_wchar_p * size.value
    raw = ref.from_address(ptr)
    args = [unicode(arg).encode("utf8") for arg in raw]
    ctypes.windll.kernel32.LocalFree(ptr)
    return args


def report(dirname, info, processCount, processTree, programList):
    dirname = os.path.abspath(dirname)
    treepath = info.get("treepath", os.path.abspath("."))
    userpath = os.path.abspath(os.path.join(
                                os.path.dirname(__file__), "userbase"))
    userpath = info.get("userpath", userpath)
    python = info.get("python", sys.executable)
    treedir = treepath.replace("\\", "/") + "/"
    userdir = userpath.replace("\\", "/") + "/Python27/site-packages/"
    pydir = os.path.dirname(python).replace("\\", "/") + "/"

    with open(os.path.join(dirname, "process-tree.json"), "w") as f:
        json.dump(processTree, f, indent=2, separators=(",",": "), sort_keys=True)

    with open(os.path.join(dirname, "program-stats.json"), "w") as f:
        json.dump(programList, f, indent=2, separators=(",",": "), sort_keys=True)

    def cputime(program):
        return program["UserTime"] + program["KernelTime"]

    programList.sort(key=cputime, reverse=True)
    programs = programList[:20]

    with open(os.path.join(dirname, "report.txt"), "w") as f:

        program = processTree["CommandLine"]
        changeset = info.get("changeset", "unknown")
        tag = info.get("tag", "")
        summary = info.get("summary")

        print >> f, ("Traced '%s' at %s (changeset %s):" %
                     (program, tag, changeset))
        print >> f
        if summary:
            print >> f, "    '%s'" % summary
            print >> f

        seconds = processTree["RealTime"] / 1000
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)

        print >> f, ("Running '%s' took %02d:%02d:%02d seconds. Ran %d processes." %
                     (program, hours, minutes, seconds, processCount))
        print >> f

        printTopPrograms(f, programs)
        for program in programs:
            if program.get("ProfileFileName"):
                printTopFunctions(f, dirname, program, treedir, pydir, userdir)


def printTopPrograms(f, programs):
    iokinds = "Read Write Other".split()
    iopsgetter = operator.itemgetter(*[k + "OperationCount" for k in iokinds])
    iobytesgetter = operator.itemgetter(*[k + "TransferCount" for k in iokinds])
    lines = []
    for program in programs:
        fields = []
        fields.append(str(program["Invocations"]))
        cputime = (program["UserTime"] + program["KernelTime"]) / 1000.0
        fields.append("%.1f s" % cputime)
        fields.append(_roundBytes(program["PeakWorkingSetSize"]))
        fields.append(str(sum(iopsgetter(program))))
        fields.append(_roundBytes(sum(iobytesgetter(program))))
        fields.append(program["Application"])
        lines.append(fields)
    print >> f, "top programs, by CPU time used:"
    print >> f
    headers = ("ncalls", "CPU time", "memory(max)",
               "IO(ops)", "IO(bytes)", "program")
    keys = range(len(headers))
    printSequences(f, keys, headers, "=", lines)
    print >> f


def _roundBytes(value):
    remainder = 0
    for unit in "B", "kB", "MB", "GB", "TB":
        if value < 1024:
            value = str(value)
            remainder = ("%.3f" % (remainder/ 1024.0))[2:-len(value)]
            if remainder:
                return "%s.%s %s" % (value, remainder, unit)
            return "%s %s" % (value, unit)
        value, remainder = divmod(value, 1024)


def printTopFunctions(f, dirname, program, treedir, pydir, userdir):
    stream = StringIO()
    path = os.path.join(dirname, program["ProfileFileName"])
    stats = pstats.Stats(path, stream=stream)
    stats.sort_stats("tottime")
    stats.print_stats(20)
    output = stream.getvalue()
    lines = output.splitlines()
    virtualenv = "/_virtualenv/"
    treedir = treedir.lower()
    userdir = userdir.lower()
    pydir = pydir.lower()
    newLines = []
    # strip header
    lines = lines[8:]
    newLines = []
    for line in lines:
        if not line:
            continue
        fields = []
        for i, field in enumerate(line.split(None, 5)):
            if i == 0:
                field = field.split("/")[0]
            elif i in (2, 4):
                continue
            elif i == 5:
                field = field.replace("\\", "/")
                tocheck = field.lower()
                if virtualenv in tocheck:
                    field = field[field.rfind(virtualenv) + len(virtualenv):]
                elif tocheck.startswith(treedir):
                    field = field[len(treedir):]
                elif tocheck.startswith(pydir):
                    field = field[len(pydir):]
                elif tocheck.startswith(userdir):
                    field = "Lib/" + field[len(userdir):]
                elif tocheck.startswith("{method"):
                    parts = field.split("'")
                    field = "{%s.%s}" % (parts[3], parts[1])
            fields.append(field)
        newLines.append(fields)

    print >> f, ("top functions for '%s', by function's internal time:" %
                 (program["Application"],))
    print >> f
    headers = "ncalls tottime cumtime filename:lineno(function)".split()
    keys = range(len(headers))
    printSequences(f, keys, headers, "-", newLines)
    print >> f


def printSequences(f, keys, headers, sep, sequences):
    lens = map(len, headers)
    lines = []
    for seq in sequences:
        fields = []
        for i, key in enumerate(keys):
            field = seq[key]
            lens[i] = max(lens[i], len(field))
            fields.append(field)
        lines.append(fields)
    paddings = tuple(lens[:-1])
    pattern = "{:>%d} " * len(paddings) + "{}"
    formatString = pattern % paddings
    seplen = len(formatString.format(*("",) * len(lens))) + lens[-1]
    print >> f, formatString.format(*headers)
    print >> f, sep * min(seplen, 78)
    for fields in lines:
        print >> f, formatString.format(*fields)


if __name__ == "__main__":
    main(*sys.argv[1:])
