
import json
import msvcrt
import os
import subprocess
import sys
import time

import win32api
import win32process
import win32profile

import mwtrace
import report



class Tracer(mwtrace.Tracer):

    # Profiling needs to be always enabled

    def prepareStartupInfo(self):
        self.started = 0
        self.reported = 0
        self.start = time.clock()
        mspdbsrv = self.mspdbsrv = _MSPDBsrv()
        self.exiting[mspdbsrv.handle] = mspdbsrv
        stdout = open(os.path.join(self.profileDir, "stdout.log"), "wb")
        stderr = open(os.path.join(self.profileDir, "stderr.log"), "wb")
        outHandle = msvcrt.get_osfhandle(stdout.fileno())
        errHandle = msvcrt.get_osfhandle(stderr.fileno())
        info = win32process.STARTUPINFO()
        info.dwFlags = win32process.STARTF_USESTDHANDLES
        info.hStdInput = self.inheritableStdHandle(win32api.STD_INPUT_HANDLE)
        info.hStdOutput = self.inheritableHandle(outHandle)
        info.hStdError = self.inheritableHandle(errHandle)
        stdout.close()
        stderr.close()
        return info


    def cleanupStartupInfo(self, info):
        self.exiting.pop(self.mspdbsrv.handle, None)
        info.hStdInput.close()
        info.hStdOutput.close()
        info.hStdError.close()


    def processStarted(self, pid):
        self.started += 1
        reported = self.started // 1000
        if reported > self.reported:
            self.reported = reported
            diff = time.clock() - self.start
            seconds = int(diff)
            minutes, seconds = divmod(seconds, 60)
            hours, minutes = divmod(minutes, 60)
            print ("%02d:%02d:%02d, seen %d processes" %
                    (hours, minutes, seconds, self.started))
            sys.stdout.flush()



class _MSPDBsrv(object):


    def __init__(self):
        self.restarts = 0
        self.starter = None
        self.handle = None
        self.restart()


    def processEnded(self):
        self.restart()


    def restart(self):
        self.restarts += 1
        if self.starter is not None:
            self.starter.wait()
            self.starter = None
            self.handle = None
        if self.restarts > 100:
            raise RuntimeError("mspdbsrv.exe restarted too many times")
        stopper = subprocess.Popen(["mspdbsrv.exe", "-stop"])
        for i in range(100):
            time.sleep(0.01)
            if stopper.poll() is not None:
                break
        else:
            raise RuntimeError("couldn't stop mspdbsrv.exe")
        starter = subprocess.Popen(["mspdbsrv.exe", "-start", "-spawn"])
        self.starter = starter
        self.handle = starter._handle



def main(outputdir, *makes):
    if not makes:
        makes = ("pymake", "mozmake")
    else:
        for make in makes:
            if make not in ("pymake", "mozmake"):
                raise RuntimeError("unknown make '%s' % make")
    info = getInfo()
    no_site_packages = os.path.join(info["treepath"],
                                         "obj-i686-pc-mingw32/_virtualenv",
                                         "Lib/no-global-site-packages.txt")
    outputdir = prepareDir(outputdir)
    for make in makes:
        if make == "pymake":
            buildargs = ["build", "-p"]
        else:
            buildargs = ["build"]
        for profile in (True, False):
            subprocess.check_call(["hg", "purge", "--all", "-X", "./.mozconfig"])
            traceMach(outputdir, info, make, profile, "configure")
            if profile:
                os.remove(no_site_packages)
            traceMach(outputdir, info, make, profile, *buildargs)


def getInfo():
    output = subprocess.check_output(["hg", "tip"])
    interesting = set("changeset tag summary".split())
    info = {}
    for line in output.splitlines():
        line = line.strip()
        if line:
            n, v = line.split(":", 1)
            if n in interesting and n not in info:
                info[n] = v.strip()
    info["python"] = os.path.abspath(sys.executable)
    info["treepath"] = os.path.abspath(".")
    info["userpath"] = os.path.abspath(os.path.join(
                                os.path.dirname(__file__), "userbase"))
    return info


def prepareDir(dirpath, throw=True):
    dirpath = os.path.abspath(dirpath)
    if os.path.exists(dirpath):
        if throw:
            raise RuntimeError("%r already exists" % dirpath)
    else:
        os.makedirs(dirpath)
    return dirpath


def traceMach(topdir, info, make, profile, *what):
    dirname = "-".join((make, what[0]))
    dirpath = prepareDir(os.path.join(topdir, dirname), profile)
    with open(os.path.join(dirpath, "info.json"), "w") as f:
        json.dump(info, f, indent=2, separators=(",",": "), sort_keys=True)
    argv = ["python", "mach"] + list(what)
    env = prepareEnvironment(info, dirpath, profile)
    tracer = Tracer(argv, env, dirpath)
    print "running '%s' with '%s'" % (" ".join(argv), make)
    processes = tracer.run()
    with open(os.path.join(dirpath, "mwtrace.json"), "w") as f:
        json.dump(processes, f, indent=2, separators=(",",": "), sort_keys=True)
    report.report(dirpath, info, *report.analyze(dirpath))


def prepareEnvironment(info, profiledir, profile):
    env = win32profile.GetEnvironmentStrings()
    exe = _abspath(info["python"])
    drive, rest = os.path.splitdrive(exe)
    python = u"/" + drive[0].lower() + rest.replace("\\", "/")
    env[u"PYTHON"] = python
    if profile:
        env.pop(u"PYTHONNOUSERSITE", None)
        env[u"PYTHONUSERBASE"] = unicode(info["userpath"])
        env[u"MWTRACE_PROFILE_DIR"] = unicode(profiledir)
    return env


def _abspath(path):
    return os.path.abspath(unicode(path, sys.getfilesystemencoding()))


if __name__ == "__main__":
    main(*sys.argv[1:])
